/*
 *	Ohio Trollius
 *	Copyright 1996 The Ohio State University
 *	RBD/NJN
 *
 *	$Id: lampack.c,v 6.1 96/11/23 22:53:07 nevin Rel $
 *
 *	Function:	- MPI data packing functions
 *			- handles recursive data types
 */

#include <mpi.h>
#include <mpisys.h>
#include <mpi_types.h>
#include <portable.h>
#include <t_types.h>
#include <typical.h>

/*
 * local functions
 */
static int		pack_hvector();
static int		pack_hindexed();
static int		pack_struct();
static int		pack_char();
static int		pack_int();
static int		pack_long();
static int		pack_short();
static int		pack_float();
static int		pack_double();
#if __STDC__
static int		pack_long_double();
#endif

/*
 *	lam_pack
 *
 *	Function:	- pack a message into contiguous space
 *			- packing is done recursively
 *			- running out of packing buffer is an error
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- packing buffer
 *			- packing buffer length
 *	Returns:	- # bytes packed or LAMERROR
 */
int
lam_pack(buf, count, dtype, pkbuf, pklen)

char			*buf;
int			count;
MPI_Datatype		dtype;
char			*pkbuf;
int			pklen;

{
	int		size;			/* # bytes packed */
/*
 * Handle the trivial cases.
 */
	if (pklen < 0) return(LAMERROR);

	if (pklen == 0) {
		return(((count == 0) || (dtype->dt_size == 0)) ? 0 : LAMERROR);
	}
/*
 * Handle the different datatypes.
 */
	switch(dtype->dt_format) {

	case LAM_DTBASIC:
		if ((dtype == MPI_UNSIGNED_CHAR) || (dtype == MPI_PACKED)
				|| (dtype == MPI_BYTE) || (dtype == MPI_CHAR)
				|| (dtype == MPI_F_CHARACTER)) {

			size = pack_char(buf, count, pkbuf, pklen);
		}
		else if ((dtype == MPI_INT) || (dtype == MPI_UNSIGNED)
				|| (dtype == MPI_F_LOGICAL)
				|| (dtype == MPI_F_INTEGER)) {

			size = pack_int(buf, count, pkbuf, pklen);
		}
		else if ((dtype == MPI_LONG)
				|| (dtype == MPI_UNSIGNED_LONG)) {

			size = pack_long(buf, count, pkbuf, pklen);
		}
		else if ((dtype == MPI_SHORT)
				|| (dtype == MPI_UNSIGNED_SHORT)) {

			size = pack_short(buf, count, pkbuf, pklen);
		}
		else if ((dtype == MPI_FLOAT) || (dtype == MPI_F_REAL)) {

			size = pack_float(buf, count, pkbuf, pklen);
		}
		else if ((dtype == MPI_DOUBLE)
				|| (dtype == MPI_F_DOUBLE_PRECISION)) {

			size = pack_double(buf, count, pkbuf, pklen);
		}
		else if (dtype == MPI_LONG_DOUBLE) {
#if __STDC__
			size = pack_long_double(buf, count, pkbuf, pklen);
#else
			size = pack_double(buf, count, pkbuf, pklen);
#endif
		}
		else if (dtype == MPI_F_COMPLEX) {

			size = pack_float(buf, 2 * count, pkbuf, pklen);
		}
		else if (dtype == MPI_F_DOUBLE_COMPLEX) {

			size = pack_double(buf, 2 * count, pkbuf, pklen);
		}
		else if ((dtype == MPI_UB) || (dtype == MPI_LB)) {

			size = 0;
		}
		else {
			size = LAMERROR;
		}

		break;

	case LAM_DTCONTIG:
		size = lam_pack(buf, count * dtype->dt_count,
					dtype->dt_dtype, pkbuf, pklen);
		break;

	case LAM_DTVECTOR:
		size = pack_hvector(buf, count, dtype, pkbuf, pklen, 1);
		break;

	case LAM_DTHVECTOR:
		size = pack_hvector(buf, count, dtype, pkbuf, pklen, 0);
		break;

	case LAM_DTINDEXED:
		size = pack_hindexed(buf, count, dtype, pkbuf, pklen, 1);
		break;

	case LAM_DTHINDEXED:
		size = pack_hindexed(buf, count, dtype, pkbuf, pklen, 0);
		break;

	case LAM_DTSTRUCT:
		size = pack_struct(buf, count, dtype, pkbuf, pklen);
		break;

	default:
		size = LAMERROR;
		break;
	}

	return(size);
}

/*
 *	pack_hvector
 *
 *	Function:	- pack vector or hvector
 */
static int
pack_hvector(frombuf, num, dtype, tobuf, tolen, fl_vec)

char			*frombuf;
int			num;
MPI_Datatype		dtype;
char			*tobuf;
int			tolen;
int			fl_vec;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes packed */
	int		totalsize;		/* total # bytes packed */
	int		extent;			/* datatype extent */
	int		stride;			/* stride in bytes */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	stride = dtype->dt_stride;
	if (fl_vec) stride *= dtype->dt_dtype->dt_upper -
				dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = frombuf;
		frombuf += extent;

		for (j = 0; j < dtype->dt_count; ++j) {

			size = lam_pack(buf, dtype->dt_length,
						dtype->dt_dtype, tobuf, tolen);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			tobuf += size;
			tolen -= size;
			buf += stride;
		}
	}

	return(totalsize);
}

/*
 *	pack_hindexed
 *
 *	Function:	- pack indexed or hindexed
 */
static int
pack_hindexed(frombuf, num, dtype, tobuf, tolen, fl_idx)

char			*frombuf;
int			num;
MPI_Datatype		dtype;
char			*tobuf;
int			tolen;
int			fl_idx;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes packed */
	int		totalsize;		/* total # bytes packed */
	int		extent;			/* datatype extent */
	int		factor;			/* extent factor */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	factor = (fl_idx == 0) ? 1 :
			dtype->dt_dtype->dt_upper - dtype->dt_dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = frombuf;
		frombuf += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp) {

			size = lam_pack(buf + (*disp * factor), *len,
						dtype->dt_dtype, tobuf, tolen);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			tobuf += size;
			tolen -= size;
		}
	}

	return(totalsize);
}

/*
 *	pack_struct
 *
 *	Function:	- pack struct
 */
static int
pack_struct(frombuf, num, dtype, tobuf, tolen)

char			*frombuf;
int			num;
MPI_Datatype		dtype;
char			*tobuf;
int			tolen;

{
	int		i, j;			/* favourite indices */
	int		size;			/* # bytes packed */
	int		totalsize;		/* total # bytes packed */
	int		extent;			/* datatype extent */
	int		*len;			/* ptr lengths */
	MPI_Aint	*disp;			/* ptr displacements */
	MPI_Datatype	*type;			/* ptr datatypes */
	char		*buf;			/* temp. buffer */

	totalsize = 0;
	extent = dtype->dt_upper - dtype->dt_lower;

	for (i = 0; i < num; ++i) {

		buf = frombuf;
		frombuf += extent;

		len = dtype->dt_lengths;
		disp = dtype->dt_disps;
		type = dtype->dt_dtypes;

		for (j = 0; j < dtype->dt_count; ++j, ++len, ++disp, ++type) {

			size = lam_pack(buf + *disp, *len,
						*type, tobuf, tolen);

			if (size < 0) return(LAMERROR);

			totalsize += size;
			tobuf += size;
			tolen -= size;
		}
	}

	return(totalsize);
}

/*
 *	pack_char
 *
 *	Function:	- pack bytes
 */
static int
pack_char(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	if (tolen < num) return(LAMERROR);

	memcpy(tobuf, frombuf, num);
	return(num);
}

/*
 *	pack_int
 *
 *	Function:	- pack integers
 */
static int
pack_int(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(int);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_int(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_int(frombuf, num);
	}

	return(size);
}

/*
 *	pack_long
 *
 *	Function:	- pack long integers
 */
static int
pack_long(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(long);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_long(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_long(frombuf, num);
	}

	return(size);
}

/*
 *	pack_short
 *
 *	Function:	- pack short integers
 */
static int
pack_short(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(short);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_short(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_short(frombuf, num);
	}

	return(size);
}

/*
 *	pack_float
 *
 *	Function:	- pack single precision floating point numbers
 */
static int
pack_float(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(float);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_float(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_float(frombuf, num);
	}

	return(size);
}

/*
 *	pack_double
 *
 *	Function:	- pack double precision floating point numbers
 */
static int
pack_double(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(double);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_double(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_double(frombuf, num);
	}

	return(size);
}

#if __STDC__
/*
 *	pack_long_double
 *
 *	Function:	- pack double precision floating point numbers
 */
static int
pack_long_double(frombuf, num, tobuf, tolen)

char			*frombuf;
int			num;
char			*tobuf;
int			tolen;

{
	int		size;			/* # bytes */

	size = num * sizeof(long double);
	if (tolen < size) return(LAMERROR);

	if (lam_homog) {
		memcpy(tobuf, frombuf, size);
	} else {
		mltot_longdbl(frombuf, num);
		memcpy(tobuf, frombuf, size);
		mttol_longdbl(frombuf, num);
	}

	return(size);
}
#endif
